; ===================================================================
; Title : BrainFuck 6502 Interpreter for the Oric
; File  : bf6502.s
; By    : iss
; ===================================================================

; ===================================================================
; Original Title: BrainFuck 6502 Interpreter for the Apple ][ //e
; Original File: BF6502A2.VER3B.S

; Platform: Apple ][ //e
; By: Michael Pohoreski
; Date: Dec, 2008
; Last updated: Jul, 2015
; Description: 187 Byte Interpreter of BrainFuck
; Version 3b
;    - No new functionality
;    - Cleaned up source code for readability
;    - Switched to Merlin directives
; License: BSD "Sharing is Caring!"
; https://github.com/Michaelangel007/brainfuck6502

; Discussion:
; https://groups.google.com/d/msg/comp.emulators.apple2/Om3JKqDZoEA/cwa5U1Hr3TAJ

; Definition:
; http://en.wikipedia.org/wiki/Brainfuck

; >  ++pData
; <  --pData
; +  ++(*pData)
; -  --(*pData)
; .  putchar(*pData)
; ,  *pData=getchar()
; [  while (*pData) { // if( *pData == 0 ), pCode = find_same_depth ( `]` )
; ]  }                // if( *pData != 0 ), pCode = find_same_depth ( `[` )

; Reference Tests:
; http://esoteric.sange.fi/brainfuck/bf-source/prog/tests.b

; Examples:
; http://esoteric.sange.fi/brainfuck/bf-source/prog/
; http://esolangs.org/wiki/Brainfuck#Implementations
; http://www.muppetlabs.com/~breadbox/bf/standards.html
; http://software.xfx.net/utilities/vbbfck/index.php
; http://nesdev.parodius.com/6502.txt

; ===================================================================

bfpc   =  $80   ; bfpc/pcode
data   =  $82   ; data/pdata

.text
        * = $9000

entry
        lda #$40        ; start code buffer = $4000..$5fff
        sta bfpc+1
        lda #$60        ; start data buffer = $6000..$7fff
        sta data+1

        lda #$00
        sta cur_depth
        sta num_bracket
        sta bfpc
        sta data
        tay

        ldx #$20        ; clear data buffer $2000 bytes
clr
        sta (data),y
        dey
        bne clr
        inc data+1
        dex
        bne clr

        lda #$60        ; start data buffer = $6000..$7fff
        sta data+1

fetch
        lda (bfpc),y
        beq exit
        jsr interpret

        inc  bfpc
        bne  fetch2
        inc  bfpc+1
fetch2
        ldy #$00
        beq fetch       ; branch always

interpret
        ldx #$07        ; 8 instructions

find_op
        cmp opcode,x    ; table of opcodes (char)
        beq exec
        dex
        bpl find_op
        rts             ; ignore non-tokens, allows for comments

exec
        lda #>entry     ; high byte of this code address
        pha
        lda opfuncptr,x ; function pointer table (address)
        pha

        clc
        lda (data),y
exit
        rts             ; 1) exit to caller,
                        ; 2) relative jsr to our bf_*(), or
                        ; 3) exit our bf_*()
bf_next
        inc   data      ; optimization: inc a3l, bne +2, inc a3h, rts
        bne   bf_next_2
        inc   data+1
bf_next_2
        rts

bf_prev
        lda data
        bne bf_prev_2
        dec data+1
bf_prev_2
        dec data

exit_2
        rts

bf_inc
        adc #$02        ; optimization: n+2-1 = n+1

bf_dec
        clc
        sbc #$00        ; optimization: fall-through into bf_incdec

store_data
        ldy #$00
        sta (data),y
        rts


bf_in
        jsr $023b
        bpl bf_in
        and #$7f        ; convert 8-bit apple text to 7-bit.ascii
        bpl store_data  ; always

bf_out
        and #$7f
        cmp #$0A
        bne do_out
        ldx #$0D
        jsr $0238
do_out
        tax
        jmp $0238


bf_if                   ; if( *pdata == 0 ) pc = `]`
        inc cur_depth   ; depth++

        lda (data),y    ; optimization: common code
        bne exit_2      ; optimization: beq _1, therefore bne rts
        lda cur_depth   ; match_depth = depth
        sta num_bracket
bf_if_2
        inc  bfpc
        bne  bf_if_3
        inc  bfpc+1
bf_if_3
        lda (bfpc),y
        cmp #'['
        bne bf_if_4
        inc num_bracket ; inc stack
        bne bf_if_2
bf_if_4
        cmp #']'
        bne bf_if_2
        lda cur_depth
        cmp num_bracket
        beq exit_2
        dec num_bracket ; dec stack
        clc
        bcc bf_if_2


bf_fi                   ; if( *pdata != 0 ) pc = `[`
        dec cur_depth   ; depth--
        lda (data),y
        beq exit_2      ; optimization: bne _1, therefore beq rts
        lda cur_depth   ; match_depth = depth
        sta num_bracket
bf_fi_2
        lda bfpc
        bne bf_fi_3
        dec bfpc+1
bf_fi_3
        dec bfpc

        lda (bfpc),y
        cmp #']'
        bne bf_fi_4
        dec num_bracket ; dec stack
        bne bf_fi_2
bf_fi_4
        cmp #'['
        bne bf_fi_2
        lda cur_depth
        cmp num_bracket
        beq exit_2
        inc num_bracket ; inc stack
        clc
        bcc bf_fi_2

cur_depth
        .byt 0          ; current nested depth

num_bracket
        .byt 0          ; depth to find[]

opcode
        .asc ",.[<]>-+" ; sorted: 2b 2c 2d 2e 3c 3e 5b 5d

opfuncptr               ; by usage: least commonly called to most
        .byt <(bf_in  -1)   ; ,
        .byt <(bf_out -1)   ; .
        .byt <(bf_if  -1)   ; [
        .byt <(bf_prev-1)   ; <
        .byt <(bf_fi  -1)   ; ]
        .byt <(bf_next-1)   ; >
        .byt <(bf_dec -1)   ; -
        .byt <(bf_inc -1)   ; +
